#
# Copyright 2015-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

import cython
cimport cython

import numpy
cimport numpy

from libc.math cimport floor, ceil
from libc.math cimport fabs
from libc.math cimport log as logc
from libc.math cimport exp as expc
# from libc.stdio cimport printf
from libcpp.vector cimport vector

ctypedef fused FType:
    double
    float
    int
    long

cdef extern from "Trace.h" namespace "Numina":
    cdef cppclass InternalTrace:
        InternalTrace() except +
        void push_back(double x, double y, double p) nogil
        double predict(double x) nogil
        vector[double] xtrace
        vector[double] ytrace
        vector[double] ptrace
        void reverse() nogil


cdef vector[int] local_max(double* mm, size_t n, double background) nogil:
    '''Find local maximum points.'''

    # TODO: this can be a pure C++ function
    
    cdef vector[int] result
    cdef size_t i
    
    if mm[0] >= background:
        if mm[0] > mm[1]:
            result.push_back(0)

    for i in range(1, n-1):
        if mm[i] >= background:
            if (mm[i] > mm[i+1]) and (mm[i] > mm[i-1]):
                result.push_back(i)
                
    if mm[n-1] >= background:
        if mm[n-1] > mm[n-2]:
            result.push_back(n-1)
            
    return result



@cython.boundscheck(False)
cdef vector[double] fit_para_equal_spaced(double y0, double y1, double y2) nogil:
    '''Fit (-1, 0, 1) (dd0, dd1, dd2) to a 2nd degree polynomial.'''

    cdef vector[double] result
    cdef double A, B, C
    C = y1
    B = 0.5 * (y2 - y0)
    A = 0.5 * (y0 + y2 - 2 * y1)

    result.push_back(A)
    result.push_back(B)
    result.push_back(C)
    return result




@cython.boundscheck(False)
cdef vector[double] interp_max_3(double y0, double y1, double y2) nogil:
    '''Parabola that passes through 3 points

    With X=[-1,0,1]
    '''

    cdef vector[double] result
    cdef vector[double] params
    cdef double A,B,C
    params = fit_para_equal_spaced(y0, y1, y2)
    A = params[0]
    B = params[1]
    C = params[2]
    if A == 0:
        result.push_back(0.0)
        result.push_back(C)
    else:
        result.push_back(-B / (2*A))
        result.push_back(C - B * B / (4*A))
    return result


cdef int wc_to_pix(double x) nogil:
    return <int>floor(x + 0.5)


@cython.cdivision(True)
@cython.boundscheck(False)
cdef int colapse_mean(FType[:, :] arr, vector[double]& out) nogil:
    cdef size_t I = arr.shape[0]
    cdef size_t J = arr.shape[1]
    cdef double accum
    cdef size_t i, j
    for i in range(I):
        accum = 0.0
        for j in range(J):
            accum += arr[i, j]
        out[i] = accum / J
    
    return 0


@cython.cdivision(True)
@cython.boundscheck(False)
cdef InternalTrace _internal_tracing(FType[:, :] arr, 
                             InternalTrace& trace, double x, double y,
                             size_t step=1, size_t hs=1, size_t tol=2,
                             double maxdis=2.0, double background=150.0,
                             int direction=-1, bint gauss = 1) nogil:

    cdef int col = wc_to_pix(x)
    cdef int row = wc_to_pix(y)
    
    cdef size_t pred_pix
    cdef double prediction
    
    cdef int axis = 1
    cdef size_t i
    cdef size_t tolcounter = tol
    cdef size_t axis_size = arr.shape[1]
    
    # Buffer
    cdef size_t regw = 1 + <int>ceil(maxdis)
    cdef size_t buffsize = 2 * regw + 1
    cdef size_t pred_off 
    cdef vector[double] pbuff
    # Peaks
    cdef vector[int] peaks
    cdef double dis, ndis
    cdef size_t ipeak
    cdef size_t nearp    
    cdef vector[double] result
    
    # Init pbuff
    for _ in range(buffsize):
        pbuff.push_back(0.0)


    while (col - step > hs) and (col + step + hs < axis_size):
        # printf("--------- %i\n", tolcounter)
        # printf("col %i dir %i step %i\n", col, direction, step)

        col += direction * step
        prediction = trace.predict(col)

        pred_pix = wc_to_pix(prediction)
        pred_off = pred_pix - regw
        # printf("col %i prediction %f pixel %i\n", col, prediction, pred_pix)

        # extract a region around the expected peak
        # and collapse it
        colapse_mean(arr[pred_pix-regw:pred_pix+regw + 1,col-hs:col+hs+1], pbuff)

        # printf("cut region %i : %i,  %i : %i\n", pred_pix-regw,pred_pix+regw,col-hs,col+hs)

        # Find the peaks
        peaks = local_max(&pbuff[0], buffsize, background=background)

        # find nearest peak to prediction
        dis = 40000.0 # a large number
        ipeak = -1
        for i in range(peaks.size()):
            ndis = fabs(peaks[i] + pred_off - prediction)
            if ndis < dis:
                dis = ndis
                ipeak = i

        # check the peak is not further than npixels'
        if ipeak < 0 or dis > maxdis:
            # printf("peak is not found %i\n", ipeak)
            # peak is not found
            if tolcounter > 0:
                # Try again
                # printf("%i tries remaining\n", tolcounter)
                tolcounter -= 1
                continue
            else:
                # No more tries
                # printf("%i tries remaining\n", 0)
                # Exit now
                return trace

        # Reset counter
        tolcounter = tol 

        nearp = peaks[ipeak] + pred_off

        # fit the peak with three points
        if gauss:
            result = interp_max_3(logc(pbuff[peaks[ipeak]-1]), logc(pbuff[peaks[ipeak]]), logc(pbuff[peaks[ipeak]+1]))
            result[1] = expc(result[1])
        else:
            result = interp_max_3(pbuff[peaks[ipeak]-1], pbuff[peaks[ipeak]], pbuff[peaks[ipeak]+1])

        if (result[0] > 0.5) or (result[0] < -0.5):
            # ignore the correction if it's larger than 0.5 pixel
            trace.push_back(col, nearp, pbuff[peaks[ipeak]])
        else:
            trace.push_back(col, result[0] + nearp, result[1])

    return trace


@cython.cdivision(True)
@cython.boundscheck(False)
def tracing(FType[:, :] arr, double x, double y, double p, size_t step=1, 
                     size_t hs=1, size_t tol=2, double background=150.0,
                     double maxdis=2.0, bint gauss = 1):
    '''Trace peak in array starting in (x,y).

    Trace a peak feature in an array starting in position (x,y).

    Parameters
    ----------
    arr : array
         A 2D array
    x : float
        x coordinate of the initial position
    y : float
        y coordinate of the initial position
    p : float
        Intensity of the initial position
    step : int, optional
           Number of pixels to move (left and rigth)
           in each iteration

    Returns
    -------
    ndarray
        A nx3 array, with x,y,p of each point in the trace
    '''
    
    cdef InternalTrace trace 
    # Initial values
    trace.push_back(x, y, p)

    _internal_tracing(arr, trace, x, y, step=step, hs=hs, tol=tol,
                      maxdis=maxdis, background=background,
                      direction=-1, gauss=gauss)
    trace.reverse()
    _internal_tracing(arr, trace, x, y, step=step, hs=hs, tol=tol,
                      maxdis=maxdis, background=background,
                      direction=+1, gauss=gauss)

    result = numpy.empty((trace.xtrace.size(), 3), dtype='float')
    for i in range(trace.xtrace.size()):
        result[i,0] = trace.xtrace[i]
        result[i,1] = trace.ytrace[i]
        result[i,2] = trace.ptrace[i]
    
    return result




#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Modify how to query results in the storage backend"""


import logging


class QueryModifier(object):
    pass


class Constraint(QueryModifier):
    pass


class ResultOf(QueryModifier):
    def __init__(self, field, node="children", ignore_fail=False,
                 id_field=None):
        from numina.types.frame import DataFrameType

        super(ResultOf, self).__init__()

        self.field = field

        if node not in ['children', 'prev', 'prev-rel', 'last']:
            raise ValueError(f"value '{node}' not allowed for node")

        self.node = node
        if self.node == 'children':
            self.id_field = id_field or "children"
        elif self.node in ['prev', 'prev-rel']:
            self.id_field = id_field or "prev"
        elif self.node == 'last':
            self.id_field = id_field or "last"

        self.ignore_fail = ignore_fail
        self.result_type = DataFrameType()

        splitm = field.split('.')
        lm = len(splitm)
        if lm == 1:
            self.mode = None
            self.attr = field
        elif lm == 2:
            self.mode = splitm[0]
            self.attr = splitm[1]
        else:
            raise ValueError(f'malformed desc: {field}')


class Ignore(QueryModifier):
    """Ignore this parameter"""
    pass


def basic_mode_builder(mode, partial_ob, backend, options=None):
    logger = logging.getLogger(__name__)

    logger.debug("builder for mode='%s'", mode.name)

    if isinstance(options, ResultOf):
        result_type = options.result_type
        name = 'relative_result'
        logger.debug('query, children id: %s', partial_ob.children)
        val = backend.search_result_relative(
            name, result_type, partial_ob, result_desc=options)
        if val is None:
            logger.debug('query, children id: %s, no result')
        for r in val:
            partial_ob.results[r.id] = r.content

    return partial_ob



#
# Copyright 2008-2017 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Module to check and activate GTC compatibility code"""

_ignore_gtc_check = False

try:
    import DF
    _run_in_gtc = True
except ImportError:
    import types
    import enum

    DF = types.ModuleType('DF')

    class AdaptedOrbTypes(enum.Enum):
        TYPE_FRAME = 1
        TYPE_STRUCT = 2
        TYPE_FRAME_LIST = 3
        TYPE_STRUCT_LIST = 4
        TYPE_DOUBLE_ARRAY2D = 5

    DF.TYPE_FRAME = AdaptedOrbTypes.TYPE_FRAME
    DF.TYPE_STRUCT = AdaptedOrbTypes.TYPE_STRUCT
    DF.TYPE_FRAME_LIST = AdaptedOrbTypes.TYPE_FRAME_LIST
    DF.TYPE_STRUCT_LIST = AdaptedOrbTypes.TYPE_STRUCT_LIST
    DF.TYPE_DOUBLE_ARRAY2D = AdaptedOrbTypes.TYPE_DOUBLE_ARRAY2D

    _run_in_gtc = False


def ignore_gtc_check(value=True):
    global _ignore_gtc_check
    _ignore_gtc_check = value


def check_gtc():
    if _ignore_gtc_check:
        return False
    else:
        return _run_in_gtc



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""


import argparse
import configparser
from importlib import import_module
import logging
import logging.config
import os
import sys

from backports.entry_points_selectable import entry_points
import importlib_resources
import yaml


from numina import __version__
from .xdgdirs import xdg_config_home
from .logconf import LOGCONF

_logger = logging.getLogger("numina")


def main(args=None):
    """Entry point for the Numina CLI."""

    # Configuration args from a text file
    config = configparser.ConfigParser()
    # Load base config here
    basecfg = importlib_resources.files().joinpath('numina.cfg')
    config.read_file(basecfg.open())

    # Extend with custom values
    read_files = config.read([
        os.path.join(xdg_config_home, 'numina/numina.cfg'),
        '.numina.cfg'
    ])
    _logger.debug(f'Reading config files {read_files}')

    parser0 = argparse.ArgumentParser(
        description='Command line interface of Numina',
        prog='numina',
        epilog="For detailed help pass --help to a target",
        add_help=False
    )

    parser0.add_argument('--disable-plugins', action='store_true')

    args0, args = parser0.parse_known_args(args)

    # Load plugin commands if enabled
    subcmd_load = []

    if not args0.disable_plugins:
        for entry in entry_points(
                group='numina.plugins.1'
        ):
            try:
                register = entry.load()
                subcmd_load.append(register)
            except Exception as error:
                print(f'exception loading plugin {entry}', file=sys.stderr)
                print(error, file=sys.stderr)

    parser = argparse.ArgumentParser(
        description='Command line interface of Numina',
        prog='numina',
        epilog="For detailed help pass --help to a target"
    )

    parser.add_argument('--disable-plugins', action='store_true',
                        help='disable plugin loading')
    parser.add_argument(
        '-l', action="store", dest="logging", metavar="FILE",
        help="FILE with logging configuration"
    )

    parser.add_argument(
        '-c', action="store", dest="config", metavar="FILE",
        help="FILE with configuration"
    )

    parser.add_argument(
        '-d', '--debug',
        action="store_true",
        dest="debug", default=False,
        help="make lots of noise"
    )

    parser.add_argument(
        '--standalone',
        action="store_true",
        dest="standalone", default=False,
        help="do not activate GTC compatibility code"
    )

    # Due to a problem with argparse
    # this command blocks the defaults of the subparsers
    # https://github.com/python/cpython/issues/53597
    # parser.set_defaults(command=None)

    subparsers = parser.add_subparsers(
        title='Targets',
        description='These are valid commands you can ask numina to do.'
    )

    # Init subcommands
    cmds = ['clidentify', 'clishowins', 'clishowom', 'clishowrecip',
            'clirun', 'clirunrec']
    for cmd in cmds:
        cmd_mod = import_module(f'.{cmd}', 'numina.user')
        register = getattr(cmd_mod, 'register', None)
        if register is not None:
            register(subparsers, config)

    # Add commands by plugins
    for register in subcmd_load:
        try:
            register(subparsers, config)
        except Exception as error:
            print(error, file=sys.stderr)

    args, unknowns = parser.parse_known_args(args)

    extra_args = process_unknown_arguments(unknowns)
    # logger file
    if args.standalone:
        import numina.ext.gtc
        numina.ext.gtc.ignore_gtc_check()

    # Config file from command line
    if args.config is not None:
        config.read_file(open(args.config))

    try:
        if args.logging is not None:
            loggingf = args.logging
        else:
            loggingf = config.get('numina', 'logging')

        with open(loggingf) as logfile:
            logconf = yaml.safe_load(logfile)
            logging.config.dictConfig(logconf)
    except configparser.Error:
        logging.config.dictConfig(LOGCONF)

    if args.debug:
        _logger.setLevel(logging.DEBUG)
        for h in _logger.handlers:
            h.setLevel(logging.DEBUG)

    _logger.info(f'Numina simple recipe runner version {__version__}')
    command = getattr(args, 'command', None)

    if command is not None:
        args.command(args, extra_args, config)


def process_unknown_arguments(unknowns):
    """Process arguments unknown to the parser"""

    result = argparse.Namespace()
    result.extra_control = {}
    # It would be interesting to use argparse internal
    # machinery for this
    for unknown in unknowns:
        # Check prefixes
        prefix = '--parameter-'
        if unknown.startswith(prefix):
            # process '='
            values = unknown.split('=')
            if len(values) == 2:
                key = values[0][len(prefix):]
                val = values[1]
                if key:
                    result.extra_control[key] = val
    return result



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

cimport cython
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np

cdef extern from "nu_ramp.h" namespace "Numina":
    cdef cppclass RampResult[T]:
        RampResult() except +
        T value
        T variance
        char npix
        char mask

    cdef cppclass HWeights:
        pass

    ctypedef vector[HWeights] HWeightsStore
    HWeightsStore create(size_t n)
    RampResult[double] axis_ramp(vector[double] buff, double dt, 
        double gain, double ron, HWeightsStore wgts_store, double blank)

@cython.boundscheck(False)
@cython.wraparound(False)
def _process_ramp_intl(datacube_t arr, double tint, double gain, double ron, mask_t badpix, double saturation, double blank,
        result_t res, 
        result_t var, 
        mask_t npix,
        mask_t mask
        ):
    cdef:
        size_t xr = arr.shape[2]
        size_t yr = arr.shape[1]
        size_t zr = arr.shape[0]
        size_t x, y, z
        RampResult[double] fres
        double val
        double dt
        char bp 
        vector[double] buff
        # weights and internal values
        vector[HWeights] wgts

    dt = tint / (zr - 1)

    buff.reserve(zr)
    wgts = create(zr)

    for x in range(xr):
        for y in range(yr):
            bp = badpix[y, x]
            if bp == MASK_GOOD:
                for z in range(zr):
                    val = arr[z,y,x]
                    if val < saturation:
                        buff.push_back(val)
                    else:
                        # We stop collecting at saturation level
                        break

                fres = axis_ramp(buff, dt, gain, ron, wgts, blank)
            else:
                fres.value = fres.variance = blank
                fres.npix = 0
                fres.mask = bp

            res[y,x] = fres.value
            var[y,x] = fres.variance
            npix[y,x] = fres.npix
            mask[y,x] = fres.mask
            buff.clear()

    return res, var, npix, mask




#
# Copyright 2018-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


class Namespace(object):
    """Namespace class, like argparse.Namespace

    >>> nm = Namespace(a=1, b="field")
    >>> nm.a == 1
    >>> nm.b == "field"

    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Check if all elements in a sequence are valid numbers (not NaN or Inf)."""
import math


def is_valid_number(x):
    """Check if x is a valid number (not NaN or Inf)."""
    return isinstance(x, (int, float)) and not math.isnan(x) and not math.isinf(x)


def all_valid_numbers(seq):
    """Check if all elements in seq are valid numbers."""
    if not isinstance(seq, (list, tuple)):
        raise TypeError("Input must be a list or tuple.")
    return all(is_valid_number(x) for x in seq)



#
# Copyright 2018-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import os.path
import json
import uuid

import pytest

from numina.util.jsonencoder import ExtEncoder
from numina.tests.drptest import create_drp_test
from ..backend import Backend
from numina.exceptions import NoResultFound
import numina.types.qc as qc
import numina.instrument.assembly as asb
import numina.core.oresult
from numina.util.context import working_directory
import numina.tests.simpleobj as simpleobj


def repeat_my(result_content, result_name):
    import numina.store

    saveres = dict(
        qc=result_content['qc'],
        uuid=result_content['uuid'],
        values={}
    )
    saveres_v = saveres['values']

    result_values = result_content['values']
    for key, obj in result_values.items():
        obj = result_values[key]
        saveres_v[key] = numina.store.dump(obj, obj, key)

    with open(result_name, 'w') as fd:
        json.dump(saveres, fd, indent=2, cls=ExtEncoder)


@pytest.fixture
def backend(tmpdir):

    drps = create_drp_test(['drpclodia.yaml'])
    name = 'CLODIA'
    ob_table = [
        dict(id=1, instrument=name, mode="sky", images=[], children=[],
             enabled=False
             ),
        dict(id=2, instrument=name, mode="sky", images=[], children=[]),
        dict(id=3, instrument=name, mode="image", images=[], children=[]),
        dict(id=4, instrument=name, mode="sky", images=[], children=[]),
        dict(id=5, instrument=name, mode="image", images=[], children=[]),
        dict(id=30, instrument=name, mode="mosaic", images=[], children=[2, 3]),
        dict(id=40, instrument=name, mode="mosaic", images=[], children=[4, 5]),
        dict(id=400, instrument=name, mode="raiz",
             images=[], children=[30, 40]),
    ]

    prod_table = {
        'TEST1': {
            '225fcaf2-7f6f-49cc-972a-70fd0aee8e96': [
                {'id': 1, 'type': 'DemoType1', 'tags': {}, 'content': {'demo1': 1}, 'ob': 2},
                {'id': 2, 'type': 'DemoType2', 'tags': {'field2': 'A'}, 'content': {'demo2': 2}, 'ob': 14},
                {'id': 3, 'type': 'DemoType2', 'tags': {'field2': 'B'}, 'content': {'demo2': 3}, 'ob': 15}
            ],
            '6ad5dc90-6b15-43b7-abb5-b07340e19f41': [
                {'id': 33, 'type': 'DemoType2', 'tags': {'field2': 'B'}, 'content': {'demo2': 9}, 'ob': 25}

            ]
        }
    }
    result_name = 'result.json'
    results_table = {
        1: {'id': 1,
            'instrument': name,
            'mode': "image",
            'oblock_id': 5,
            'task_id': 1,
            'time_create': '2018-07-24T19:12:01',
            'result_dir': 'dum1',
            'result_file': result_name,
            'qc': 'GOOD', },
        2: {'id': 2,
            'instrument': name,
            'mode': "sky",
            'oblock_id': 4,
            'task_id': 2,
            'qc': 'BAD',
            'time_create': '2018-07-24T19:12:09',
            'result_file': result_name,
            'result_dir': 'dum2', },
        3: {'id': 3,
            'instrument': name,
            'mode': "image",
            'oblock_id': 5,
            'task_id': 3,
            'time_create': '2018-07-24T19:12:11',
            'qc': 'GOOD',
            'result_dir': 'dum3',
            'result_file': result_name, },
    }

    gentable = {}
    gentable['products'] = prod_table
    gentable['requirements'] = {}
    gentable['results'] = results_table
    # gentable['oblocks'] = ob_table
    # Load instrument profiles
    pkg_paths = ['numina.drps.tests.configs']
    store = asb.load_paths_store(pkg_paths)

    result_name = 'result.json'

    result1_dir = tmpdir.mkdir('dum1')
    result1_values = dict(
        calib=simpleobj.create_simple_structured(),
        reduced_rss=simpleobj.create_simple_frame(),
        reduced_image=simpleobj.create_simple_frame()
    )
    result1_content = dict(
        qc='GOOD',
        values=result1_values,
        uuid='10000000-10000000-10000000-10000000'
    )
    with working_directory(str(result1_dir)):
        repeat_my(result1_content, result_name)

    result2_dir = tmpdir.mkdir('dum2')

    result2_values = dict(calib=simpleobj.create_simple_structured())
    result2_content = dict(
        qc='BAD',
        values=result2_values,
        uuid='20000000-20000000-20000000-20000000'
    )
    with working_directory(str(result2_dir)):
        repeat_my(result2_content, result_name)

    result3_dir = tmpdir.mkdir('dum3')

    result3_values = dict(
        reduced_rss='reduced_rss.fits',
        reduced_image='reduced_image.fits'
    )
    result3_content = dict(
        qc='BAD',
        values=result3_values,
        uuid='30000000-30000000-30000000-30000000'
    )
    with working_directory(str(result3_dir)):
        repeat_my(result3_content, result_name)

    base = Backend(drps, gentable, components=store, basedir=str(tmpdir))
    base.add_obs(ob_table)

    return base


@pytest.fixture
def backend_empty(tmpdir):
    drps = create_drp_test(['drpclodia.yaml'])
    pkg_paths = ['numina.drps.tests.configs']
    store = asb.load_paths_store(pkg_paths)
    base = Backend(drps, {}, components=store, basedir=str(tmpdir))
    return base


def test_skip_reserved(backend):

    ss_ids = list(backend.search_session_ids())

    assert ss_ids == [2, 3, 4, 5, 30, 40]


def test_parent_inserted(backend):

    obsres = backend.search_oblock_from_id(2)
    assert obsres.parent == 30

    obsres = backend.search_oblock_from_id(4)
    assert obsres.parent == 40

    obsres = backend.search_oblock_from_id(30)
    assert obsres.parent == 400

    obsres = backend.search_oblock_from_id(400)
    assert obsres.parent is None


def test_previous_obsid(backend):

    obsres = backend.search_oblock_from_id(5)
    print(type(obsres))
    previd = backend.search_previous_obsres(obsres, node='prev')
    assert list(previd) == [4, 3, 2, 1]

    obsres = backend.search_oblock_from_id(5)
    previd = backend.search_previous_obsres(obsres, node='prev-rel')
    assert list(previd) == [4]

    obsres = backend.search_oblock_from_id(4)
    previd = backend.search_previous_obsres(obsres, node='prev-rel')
    assert list(previd) == []


def test_search_result_id(backend):
    from numina.types.frame import DataFrameType
    from numina.types.dataframe import DataFrame

    tipo = DataFrameType()
    node_id = 5
    field = 'reduced_rss'
    res = backend.search_result_id(node_id, tipo, field, mode=None)

    assert isinstance(res.content, DataFrame)
    relpath = os.path.relpath(res.content.filename, backend.basedir)
    assert relpath == os.path.join('dum3', 'reduced_rss.fits')


def test_search_result_id_notfound(backend):
    from numina.types.frame import DataFrameType

    tipo = DataFrameType()
    node_id = 2
    field = 'reduced_rss'

    with pytest.raises(NoResultFound):
        backend.search_result_id(node_id, tipo, field, mode=None)


def test_build_recipe_result(backend):
    from numina.types.dataframe import DataFrame
    from numina.types.structured import BaseStructuredCalibration

    obj_uuid = '10000000-10000000-10000000-10000000'
    res = backend.build_recipe_result(result_id=1)

    assert res.qc == qc.QC.GOOD

    assert hasattr(res, 'reduced_rss')
    assert isinstance(res.reduced_rss, DataFrame)

    assert hasattr(res, 'reduced_image')
    assert isinstance(res.reduced_image, DataFrame)

    assert hasattr(res, 'calib')
    assert isinstance(res.calib, BaseStructuredCalibration)

    assert res.uuid == uuid.UUID(obj_uuid)


def test_build_recipe_result2(backend):
    from numina.types.structured import BaseStructuredCalibration
    obj_uuid = '20000000-20000000-20000000-20000000'
    res = backend.build_recipe_result(result_id=2)

    assert res.qc == qc.QC.BAD

    assert hasattr(res, 'calib')
    assert isinstance(res.calib, BaseStructuredCalibration)

    assert res.uuid == uuid.UUID(obj_uuid)


def test_ago(backend_empty):

    ob1 = {'id': 100, 'instrument': 'CLODIA',
           'mode': 'image', 'labels': {'obsid_wl': 400}}
    ob2 = {'id': 200, 'instrument': 'CLODIA',
           'mode': 'image', 'labels': {'obsid_wl': 400}}
    ob3 = {'id': 300, 'instrument': 'CLODIA', 'mode': 'image'}
    ob4 = {'id': 400, 'instrument': 'CLODIA', 'mode': 'image'}

    backend_empty.add_obs([ob1, ob2, ob3, ob4])

    oblock1 = backend_empty.oblock_from_id(100)
    assert isinstance(oblock1, numina.core.oresult.ObservingBlock)

    assert oblock1.labels == ob1['labels']

    oblock2 = backend_empty.oblock_from_id(200)
    assert isinstance(oblock2, numina.core.oresult.ObservingBlock)

    obsres1 = backend_empty.obsres_from_oblock(oblock1)

    assert isinstance(obsres1, numina.core.oresult.ObservationResult)



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""User command line interface of Numina."""

import datetime
import errno
import logging
import os
import pickle
import pkgutil
import shutil

import yaml

import numina.drps
from numina.dal.backend import Backend
from numina.dal.dictdal import HybridDAL
from numina.util.jsonencoder import ExtEncoder
from numina.types.frame import DataFrameType
from numina.types.qc import QC
from numina.util.context import working_directory


_logger = logging.getLogger(__name__)


class DataManager:
    def __init__(self, basedir, datadir, backend):
        self.basedir = basedir
        self.datadir = datadir
        self.backend = backend

        self.workdir_tmpl = "obsid{obsid}_{taskid}_work"
        self.resultdir_tmpl = "obsid{obsid}_{taskid}_result"
        self.serial_format = 'json'

        self.resultfile_tmpl = 'result.json'
        self.taskfile_tmpl = 'task.json'

    def insert_obs(self, loaded_obs):
        self.backend.add_obs(loaded_obs)

    def load_observations(self, obfile):

        loaded_obs = []
        with open(obfile) as fd:
            sess = []
            for doc in yaml.safe_load_all(fd):
                enabled = doc.get('enabled', True)
                docid = doc['id']
                requirements = doc.get('requirements', {})
                sess.append(
                    dict(id=docid, enabled=enabled,
                         requirements=requirements)
                )
                loaded_obs.append(doc)
        self.insert_obs(loaded_obs)
        return loaded_obs

    def serializer(self, data, fd):
        if self.serial_format == 'yaml':
            self.serializer_yaml(data, fd)
        elif self.serial_format == 'json':
            self.serializer_json(data, fd)
        else:
            raise ValueError('serializer not supported')

    def serializer_json(self, data, fd):
        import json
        json.dump(data, fd, indent=2, cls=ExtEncoder)

    def serializer_yaml(self, data, fd):
        import yaml
        yaml.dump(data, fd)

    def store_result_to(self, result):
        saveres = result.store_to(None)
        return saveres

    def store_task(self, task):

        result_dir_rel = task.request_runinfo['results_dir']
        result_dir = os.path.join(self.basedir, result_dir_rel)

        values = dict(
            obsid=task.request_params['oblock_id'],
            taskid=task.id
        )

        result_file = self.resultfile_tmpl.format(**values)
        task_file = self.taskfile_tmpl.format(**values)

        with working_directory(result_dir):

            task_repr = task.__dict__.copy()

            # save to disk the RecipeResult part and return the file to save it
            if task.result is not None:
                uuid = task.result.uuid
                qc = task.result.qc
                _logger.info('storing result uuid=%s, quality=%s', uuid, qc)
                if qc != QC.GOOD:
                    for key, val in task.result.stored().items():
                        val = getattr(task.result, key)
                        if hasattr(val, 'quality_control'):
                            _logger.info('with field %s=%s, quality=%s',
                                         key, val, val.quality_control)

                result_repr = self.store_result_to(task.result)
                # Change result structure by filename
                task_repr['result'] = result_file

                with open(result_file, 'w+') as fd:
                    self.serializer(result_repr, fd)

            tid = task.id
            _logger.info('storing task id=%s', tid)
            with open(task_file, 'w+') as fd:
                self.serializer(task_repr, fd)

        self.backend.update_task(task)
        if task.result is not None:
            self.backend.update_result(task, result_repr, result_file)

    def create_workenv(self, task):

        values = dict(
            obsid=task.request_params['oblock_id'],
            taskid=task.id
        )

        work_dir = self.workdir_tmpl.format(**values)
        result_dir = self.resultdir_tmpl.format(**values)

        workenv = WorkEnvironment(
            self.datadir,
            self.basedir,
            work_dir,
            result_dir
        )

        return workenv


class ProcessingTask:
    def __init__(self):

        self.result = None
        self.id = 1

        self.time_create = datetime.datetime.now(datetime.timezone.utc)
        self.time_start = 0
        self.time_end = 0
        self.request = "reduce"
        self.request_params = {}
        self.request_runinfo = self._init_runinfo()
        self.state = 0

    @classmethod
    def _init_runinfo(cls):
        request_runinfo = dict(runner='unknown', runner_version='unknown')
        return request_runinfo


class WorkEnvironment:
    def __init__(self, datadir, basedir, workdir,
                 resultsdir):

        self.basedir = basedir

        self.workdir_rel = workdir
        self.workdir = os.path.abspath(os.path.join(self.basedir, workdir))

        self.resultsdir_rel = resultsdir
        self.resultsdir = os.path.abspath(
            os.path.join(self.basedir, resultsdir))

        self.datadir_rel = datadir
        self.datadir = os.path.abspath(datadir)

        index_base = "index.pkl"

        self.index_file = os.path.join(self.workdir, index_base)
        self.hashes = {}

    def sane_work(self):
        _logger.debug('check workdir for working: %r', self.workdir_rel)
        make_sure_path_exists(self.workdir)
        make_sure_file_exists(self.index_file)
        # Load dictionary of hashes

        with open(self.index_file, 'rb') as fd:
            try:
                self.hashes = pickle.load(fd)
            except EOFError:
                self.hashes = {}
        # make_sure_path_doesnot_exist(self.resultsdir)
        make_sure_file_exists(self.index_file)

        # make_sure_path_doesnot_exist(self.resultsdir)
        _logger.debug('check resultsdir to store results %r',
                      self.resultsdir_rel)
        make_sure_path_exists(self.resultsdir)

    def copyfiles(self, obsres, reqs):

        _logger.info('copying files from %r to %r',
                     self.datadir_rel, self.workdir_rel)

        if obsres:
            self.copyfiles_stage1(obsres)

        self.copyfiles_stage2(reqs)

    def installfiles_stage1(self, obsres, action='copy'):
        import astropy.io.fits as fits

        install_if_needed = self._calc_install_if_needed(action)

        _logger.debug('installing files from observation result')
        tails = []
        sources = []
        for f in obsres.images:
            if not os.path.isabs(f.filename):
                complete = os.path.abspath(
                    os.path.join(self.datadir, f.filename))
            else:
                complete = f.filename
            head, tail = os.path.split(complete)
            # initial.append(complete)
            tails.append(tail)
            #            heads.append(head)
            sources.append(complete)

        dupes = self.check_duplicates(tails)

        for src, obj in zip(sources, obsres.images):
            head, tail = os.path.split(src)
            if tail in dupes:
                # extract UUID
                hdr = fits.getheader(src)
                img_uuid = hdr['UUID']
                root, ext = os.path.splitext(tail)
                key = f"{root}_{img_uuid}{ext}"

            else:
                key = tail
            dest = os.path.join(self.workdir, key)
            # Update filename in DataFrame
            obj.filename = dest
            install_if_needed(key, src, dest)

        if obsres.results:
            _logger.warning("not installing files in 'results")
        return obsres

    def _calc_install_if_needed(self, action):
        if action not in ['copy', 'link']:  # , 'symlink', 'hardlink']:
            raise ValueError(f"{action} action is not allowed")

        _logger.debug(f'installing files with "{action}"')

        if action == 'copy':
            install_if_needed = self.copy_if_needed
        elif action == 'link':
            install_if_needed = self.link_if_needed
        else:
            raise ValueError(f"{action} action is not allowed")
        return install_if_needed

    def installfiles_stage2(self, reqs, action='copy'):
        _logger.debug('installing files from requirements')

        install_if_needed = self._calc_install_if_needed(action)

        for _, req in reqs.stored().items():
            if isinstance(req.type, DataFrameType):
                value = getattr(reqs, req.dest)
                if value is None:
                    continue

                complete = os.path.abspath(
                    os.path.join(self.datadir, value.filename)
                )

                head, tail = os.path.split(value.filename)
                dest = os.path.join(self.workdir, tail)

                install_if_needed(value.filename, complete, dest)

    def copyfiles_stage1(self, obsres):
        return self.installfiles_stage1(obsres, action='copy')

    def linkfiles_stage1(self, obsres):
        return self.installfiles_stage1(obsres, action='link')

    def check_duplicates(self, tails):
        seen = set()
        dupes = set()
        for tail in tails:
            if tail in seen:
                dupes.add(tail)
            else:
                seen.add(tail)
        return dupes

    def copyfiles_stage2(self, reqs):
        return self.installfiles_stage2(reqs, action='copy')

    def linkfiles_stage2(self, reqs):
        return self.installfiles_stage2(reqs, action='link')

    def copy_if_needed(self, key, src, dest):

        md5hash = compute_md5sum_file(src)
        _logger.debug('compute hash, %s %s %s', key, md5hash, src)

        # Check hash
        hash_in_file = self.hashes.get(key)
        if hash_in_file is None:
            trigger_save = True
            make_copy = True
        elif hash_in_file == md5hash:
            trigger_save = False
            if os.path.isfile(dest):
                make_copy = False
            else:
                make_copy = True
        else:
            trigger_save = True
            make_copy = True

        self.hashes[key] = md5hash

        if make_copy:
            _logger.debug('copying %r to %r', key, self.workdir)
            shutil.copy(src, dest)
        else:
            _logger.debug('copying %r not needed', key)

        if trigger_save:
            _logger.debug('save hashes')
            with open(self.index_file, 'wb') as fd:
                pickle.dump(self.hashes, fd)

    def link_if_needed(self, key, src, dest):
        _logger.debug('linking %r to %r', key, self.workdir)
        try:
            # Remove destination
            os.remove(dest)
        except OSError:
            pass
        os.symlink(src, dest)

    def adapt_obsres(self, obsres):
        """Adapt obsres after file copy"""

        _logger.debug('adapt observation result for work dir')
        for f in obsres.images:
            # Remove path components
            f.filename = os.path.basename(f.filename)
        return obsres


def compute_md5sum_file(filename):
    import hashlib
    md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(128 * md5.block_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def make_sure_path_doesnot_exist(path):
    try:
        shutil.rmtree(path)
    except (OSError, IOError) as exception:
        if exception.errno != errno.ENOENT:
            raise


def make_sure_path_exists(path):
    try:
        os.makedirs(path)
    except (OSError, IOError) as exception:
        if exception.errno != errno.EEXIST:
            raise


def make_sure_file_exists(path):
    try:
        with open(path, 'a'):
            pass
    except (OSError, IOError) as exception:
        if exception.errno != errno.EEXIST:
            raise


def process_format_version_1(basedir, loaded_data, loaded_data_extra=None, profile_path_extra=None) -> HybridDAL:
    import numina.instrument.assembly as asbl
    sys_drps = numina.drps.get_system_drps()
    com_store = asbl.load_panoply_store(sys_drps, profile_path_extra)
    backend = HybridDAL(
        sys_drps, [], loaded_data,
        extra_data=loaded_data_extra,
        basedir=basedir,
        components=com_store
    )
    return backend


def process_format_version_2(basedir, loaded_data, loaded_data_extra=None,
                             profile_path_extra=None, filename=None) -> Backend:
    import numina.instrument.assembly as asbl
    sys_drps = numina.drps.get_system_drps()
    com_store = asbl.load_panoply_store(sys_drps, profile_path_extra)
    loaded_db = loaded_data['database']
    backend = Backend(
        sys_drps, loaded_db,
        extra_data=loaded_data_extra,
        basedir=basedir,
        components=com_store,
        filename=filename
    )
    backend.rootdir = loaded_data.get('rootdir', '')
    return backend


def create_datamanager(config, reqfile,
                       extra_control=None, profile_path_extra=None,
                       persist=True) -> DataManager:

    section = config['tool.run']
    basedir = section['basedir']
    datadir = section['datadir']

    if reqfile:
        _logger.info('reading control from %s', reqfile)
        with open(reqfile, 'r') as fd:
            loaded_data = yaml.safe_load(fd)
    else:
        _logger.info('no control file')
        loaded_data = {}

    if extra_control:
        _logger.info('extra control %s', extra_control)
        loaded_data_extra = parse_as_yaml(extra_control)
    else:
        loaded_data_extra = None

    control_format = loaded_data.get('version', 1)
    _logger.info('control format version %d', control_format)

    if control_format == 1:
        _backend = process_format_version_1(
            basedir, loaded_data, loaded_data_extra, profile_path_extra)
        datamanager = DataManager(basedir, datadir, _backend)
        datamanager.workdir_tmpl = section['workdir_tmpl']
        datamanager.resultdir_tmpl = section['resultdir_tmpl']
        datamanager.resultfile_tmpl = section['resultfile_tmpl']
        datamanager.taskfile_tmpl = section['taskfile_tmpl']
    elif control_format == 2:
        if persist:
            pname = reqfile
        else:
            pname = None
        _backend = process_format_version_2(basedir, loaded_data,
                                            loaded_data_extra,
                                            profile_path_extra,
                                            filename=pname)

        datamanager = DataManager(basedir, datadir, _backend)
    else:
        msg = f'Unsupported format {control_format} in {reqfile}'
        raise ValueError(msg)

    # This should go before we load CL file
    # load additional reduction defaults
    for ins, drp in datamanager.backend.drps.query_all().items():
        pkg = f"{drp.package}.recipes"
        resource = "configs.yaml"
        try:
            data = pkgutil.get_data(pkg, resource)
            values = yaml.safe_load(data)
            # insert requirements
            reqs = values.get('requirements', {})
            reqs_ins = reqs.get(ins, {})
            for prof_name in reqs_ins:
                node_pln = reqs_ins[prof_name]
                for pln_name in node_pln:
                    node_obs = node_pln[pln_name]
                    for obs_name in node_obs:
                        params = node_obs[obs_name]
                        # Set instrument if not defined
                        n1 = datamanager.backend.req_table.setdefault(ins, {})
                        # Set instrumental profile if not defined
                        n2 = n1.setdefault(prof_name, {})
                        # Set pipeline if not defined
                        n3 = n2.setdefault(pln_name, {})
                        # Insert params
                        n3[obs_name] = params

        except FileNotFoundError:
            # if the file doesn't exist, we ignore it
            pass

    return datamanager


def load_observations(obfiles, is_session=False):
    """Observing mode processing mode of numina."""

    # Loading observation result if exists
    loaded_obs = []
    sessions = []

    for obfile in obfiles:
        with open(obfile) as fd:
            if is_session:
                _logger.info("session file from %r", obfile)
                sess = yaml.safe_load(fd)
                sessions.append(sess['session'])
            else:
                _logger.info("observation results from %r", obfile)
                sess = []
                for doc in yaml.safe_load_all(fd):
                    enabled = doc.get('enabled', True)
                    docid = doc['id']
                    requirements = doc.get('requirements', {})
                    sess.append(dict(id=docid, enabled=enabled,
                                     requirements=requirements))
                    if enabled:
                        _logger.debug(
                            "load observation result with id %s", docid)
                    else:
                        _logger.debug(
                            "skip observation result with id %s", docid)

                    loaded_obs.append(doc)

            sessions.append(sess)
    return sessions, loaded_obs


def parse_as_yaml(strdict):
    """Parse a dictionary of strings as if yaml reads it"""
    interm = ""
    for key, val in strdict.items():
        interm = f"{key}: {val}, {interm}"
    fin = '{%s}' % interm

    return yaml.safe_load(fin)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Utility classes and functions to employ R from python using rpy2."""

import numpy as np
from scipy.special import comb
import rpy2.robjects as ro
from rpy2.robjects.packages import importr


class LinearModelYvsX:
    """Linear Model class to fit data using the R lm() function.

    Parameters
    ----------
    x : 1d numpy array, float
        X values.
    y : 1d numpy array, float
        Y values.
    degree: int
        Polynomial degree.
    w : 1d numpy array, float
        Optional vector of weights to be used in the fitting process.
    raw : bool
        If True, use raw and not orthogonal polynomials.
    renorm : bool
        If True, renormalize the data ranges to [-1,1] before the fit
        as explained in Appendix B1 of Cardiel (2009, MNRAS, 396, 680)
        in order to prevent numerical errors.
    rm_all : bool
        If True, remove all R objects to initialize session.
    debug : bool
        If True, intermediate information is displayed.

    Attributes
    ----------
    fit : rpy2.robjects.vectors.ListVector
        Object returned by call to lm(...) function.
    summary : rpy2.robjects.vectors.ListVector
        Object returned by call to summary(lm(...))) function.
    coeff_estimate : 1d numpy array, float
        Fitted coefficients.
    coeff_std_error: 1d numpy array, float
        Associated standard deviation for the coefficients.
    coeff_t_value : 1d numpy array, float
        t value returned by summary(lm(...))
    coeff_wald_test : 1d numpy array, float
        Pr(>|t|) returned by summary(lm(...))
    p_value : float
        Global p-value obtained in the F-test.

    Examples
    --------
    Simple fit to a 5th order polynomial

    >>> import numpy as np
    >>> x = np.arange(10)
    >>> y = 2 * x
    >>> w = np.ones(x.size)
    >>> myfit = LinearModelYvsX(x=x, y=y, degree=5, w=w)
    >>> print(myfit.coeff_estimate)
    [  1.12346671e-15   2.00000000e+00   8.55144851e-16  -1.17227862e-16
      -2.12064864e-19   5.50598230e-19]
    >>> print(myfit.coeff_std_error)
    [  6.28322594e-16   1.74104426e-15   1.37908919e-15   4.10725784e-16
       5.12878797e-17   2.26887586e-18]

    Note that relevant information of the resulting fit is returned in
    the class member 'fit'. Individual values are accessible using the
    method rx2.

    >>> print(myfit.fit.names)
     [1] "coefficients"  "residuals"     "fitted.values" "effects"
     [5] "weights"       "rank"          "assign"        "qr"
     [9] "df.residual"   "xlevels"       "call"          "terms"
    [13] "model"
    >>> print(myfit.fit.rx2("coefficients"))

    In addition, the result of the call to the R function
    summary(lm(...)) is also stored as the class member 'summary'.
    Individual values are also accessible using the method rx2.

    >>> print(myfit.summary.names)
     [1] "call"          "terms"         "weights"       "residuals"
     [5] "coefficients"  "aliased"       "sigma"         "df"
     [9] "r.squared"     "adj.r.squared" "fstatistic"    "cov.unscaled"
    >>> print(myfit.summary.rx2("sigma"))
    [1] 6.336625e-16

    """

    def __init__(self, x, y, degree, w=None,
                 raw=True, renorm=False,
                 rm_all=False, debug=False):
        """Fit polynomial using the R lm() function."""

        # remove all objects in R session
        if rm_all:
            ro.r('rm(list=ls())')

        # renormalize data ranges when requested
        if renorm:
            xmin = x.min()
            xmax = x.max()
            ymin = y.min()
            ymax = y.max()
            if xmin == xmax:
                bx = 1.0
                cx = 0.0
            else:
                bx = 2./(xmax - xmin)
                cx = (xmax + xmin)/(xmax - xmin)
            if ymin == ymax:
                by = 1.0
                cy = 0.0
            else:
                by = 2./(ymax - ymin)
                cy = (ymax + ymin)/(ymax - ymin)
            xx = bx * x - cx
            yy = by * y - cy
            if w is None:
                ww = None
            else:
                ww = by * w - cy
        else:
            xx = np.copy(x)
            yy = np.copy(y)
            if w is None:
                ww = None
            else:
                ww = np.copy(w)

        # declare x and y in R session
        ro.globalenv['x'] = ro.FloatVector(xx)
        ro.globalenv['y'] = ro.FloatVector(yy)

        # define R command line to be executed
        r_command = 'lm(y ~ poly(x, degree=' + str(degree)
        r_command += ', raw=' + str(raw).upper() + ')'
        if w is not None:  # weigthed fit
            # declare weights in R session
            ro.globalenv['w'] = ro.FloatVector(ww)
            r_command += ', weights=w'
        r_command += ')'

        if debug:
            print('r_command:\n', r_command)

        # execute command line and store resulting fit
        self.fit = ro.r(r_command)

        # store summary
        base = importr('base')
        self.summary = base.summary(self.fit)

        # store coefficients and associated statistics
        coeff_matrix = np.array(self.summary.rx2("coefficients"))
        self.coeff_estimate = coeff_matrix[:, 0]
        self.coeff_std_error = coeff_matrix[:, 1]
        self.coeff_t_value = coeff_matrix[:, 2]
        self.coeff_wald_test = coeff_matrix[:, 3]

        # when necessary, reverse the impact of renormalization of
        # the data ranges in the fitted coefficients
        if renorm:
            if cx == 0.0:
                coeff_estimate = np.copy(self.coeff_estimate)
                for i in range(degree + 1):
                    coeff_estimate[i] *= bx**i
            else:
                aa = np.copy(self.coeff_estimate)
                coeff_estimate = np.zeros_like(self.coeff_estimate)
                # coeff_std_error = np.zeros_like(self.coeff_std_error)
                for i in range(degree + 1):
                    for j in range(i, degree + 1):
                        coeff_estimate[i] += aa[j] * comb(j, j-i, True) * \
                            (bx**i)*((-cx)**(j-i))
            coeff_estimate[0] += cy
            coeff_estimate /= by
            self.coeff_estimate = coeff_estimate

        # compute global p-value, which is not stored in summary
        fstatistic = self.summary.rx2("fstatistic")
        stats = importr('stats')
        global_p_value = stats.pf(q=fstatistic[0],
                                  df1=fstatistic[1],
                                  df2=fstatistic[2],
                                  lower_tail=False)
        self.p_value = np.array(global_p_value)[0]



#
# Copyright 2016-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import traceback
import collections


class Signal:
    """Signal used for callbacks."""

    def __init__(self):
        self.callbacks = collections.OrderedDict()
        self._cid_counter = 0

    def connect(self, callback):
        self._cid_counter += 1
        cid = self._cid_counter
        self.callbacks[cid] = callback
        return cid

    def delete(self, idx):
        del self.callbacks[idx]

    def emit(self, *args, **kwds):
        result = []
        for cid, cb in self.callbacks.items():
            try:
                res = cb(*args, **kwds)
                result.append((cid, res))
            except TypeError:
                traceback.print_exc()
        return result



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import contextlib
import enum
from typing import Generator


from .signal import Signal


class Status(enum.Enum):
    """Enum representing the status of the component"""

    ACTIVE = 1
    FAILED = 2
    RESETTING = 3
    CONFIGURING = 4


class Transition(enum.Enum):
    RESET = 0
    END_RESET = 1
    CONFIGURE = 2
    END_CONFIGURE = 3
    ERROR = 4


class State:
    def __init__(self):
        self.current = Status.ACTIVE
        #
        self.enter_configure = Signal()
        self.exit_configure = Signal()
        self.enter_reset = Signal()
        self.exit_reset = Signal()
        self.enter_active = Signal()
        self.exit_active = Signal()
        self.enter_failed = Signal()
        self.exit_failed = Signal()

    @property
    def is_configuring(self) -> bool:
        return self.current == Status.CONFIGURING

    def transition(self, inpt: Transition) -> None:
        match self.current:
            case Status.FAILED:
                match inpt:
                    case Transition.RESET:
                        self.exit_failed.emit()
                        self.current = Status.RESETTING
                        self.enter_reset.emit()
                    case _:
                        pass  # No action
            case Status.RESETTING:
                match inpt:
                    case Transition.RESET:
                        pass  # No action
                    case Transition.END_RESET:
                        self.exit_reset.emit()
                        self.current = Status.ACTIVE
                        self.enter_active.emit()
                    case _:
                        self.exit_reset.emit()
                        self.current = Status.FAILED
                        self.enter_failed.emit()
            case Status.CONFIGURING:
                match inpt:
                    case Transition.CONFIGURE:
                        pass  # No action
                    case Transition.END_CONFIGURE:
                        self.current = Status.ACTIVE
                        self.exit_configure.emit()
                    case _:
                        self.current = Status.FAILED
                        self.enter_failed.emit()
            case Status.ACTIVE:
                match inpt:
                    case Transition.CONFIGURE:
                        self.current = Status.CONFIGURING
                        self.enter_configure.emit()
                    case Transition.RESET:
                        self.current = Status.RESETTING
                        self.enter_reset.emit()
                    case _:
                        self.current = Status.FAILED
                        self.enter_failed.emit()

    @contextlib.contextmanager
    def managed_configure(self) -> Generator["State", None, None]:
        self.transition(Transition.CONFIGURE)
        try:
            yield self
        finally:
            self.transition(Transition.END_CONFIGURE)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Auxiliary script to perform binary image arithmetic with 3D cubes.

The computation is performed casting to np.float32."""

import argparse
import sys

from astropy.io import fits
import numpy as np
import pathlib

from .add_script_info_to_fits_history import add_script_info_to_fits_history

def compute_operation(file1, file2, operation, extname1, extname2, dtype):
    """Compute output = file1 operation file2.

    Parameters
    ----------
    file1 : file object
        First FITS file.
    file2 : file object or float
        Second FITS file or float number.
    operation : string
        Mathematical operation.
    extname1 : str
        Extension name of the first FITS file (default: 'PRIMARY').
    extname2 : str
        Extension name of the second FITS file (default: 'PRIMARY').
    dtype : str
        Data type of the output image.

    Returns
    -------
    solution : `numpy.ndarray`
        Resulting 3D image after applying the operation.
    """

    # read first FITS file
    with fits.open(file1) as hdulist:
        image_header1 = hdulist[extname1].header
        image1 = hdulist[extname1].data.astype(dtype)
    naxis = image_header1['naxis']
    if naxis != 3:
        raise ValueError("Input file must be a 3D cube (NAXIS=3).")
    naxis1 = image_header1['naxis1']
    naxis2 = image_header1['naxis2']
    naxis3 = image_header1['naxis3']

    # read second FITS file or number
    try:
        with fits.open(file2) as hdulist:
            image_header2 = hdulist[extname2].header
            image2 = hdulist[extname2].data.astype(dtype)
            naxis_ = image_header2['naxis']
            naxis1_ = image_header2['naxis1']
            naxis2_ = image_header2['naxis2']
            naxis3_ = image_header2['naxis3']
    except FileNotFoundError:
        image2 = np.zeros((naxis3, naxis2, naxis1), dtype=dtype)
        if 'int' in dtype:
            file2 = int(file2)
        elif 'float' in dtype:
            file2 = float(file2)
        else:
            raise ValueError(f"Unsupported dtype: {dtype}. Use an integer or float type.")
        image2 += file2  # if file2 is a number
        naxis_ = naxis
        naxis1_ = naxis1
        naxis2_ = naxis2
        naxis3_ = naxis3

    # additional dimension checks
    if naxis != naxis_:
        raise ValueError("NAXIS values are different.")
    if naxis1 != naxis1_:
        raise ValueError("NAXIS1 values are different.")
    if naxis2 != naxis2_:
        raise ValueError("NAXIS2 values are different.")
    if naxis3 != naxis3_:
        raise ValueError("NAXIS3 values are different.")

    # compute operation
    if operation == "+":
        solution = image1 + image2
    elif operation == "-":
        solution = image1 - image2
    elif operation == "x":
        solution = image1 * image2
    elif operation == "/":
        solution = image1 / image2
    elif operation == "=":
        solution = image2.copy()
    else:
        raise ValueError("Unexpected operation=" + str(operation))

    return solution


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description="description: binary image arithmetic"
    )
    # positional parameters
    parser.add_argument("file1",
                        help="First FITS image",
                        type=argparse.FileType('rb'))
    parser.add_argument("operation",
                        help="Arithmetic operation",
                        type=str,
                        choices=['+', '-', 'x', '/', '='])
    parser.add_argument("file2",
                        help="Second FITS image or number",
                        type=str)
    parser.add_argument("output",
                        help="Output FITS image",
                        type=str)
    # optional arguments
    parser.add_argument("--extname1", type=str,
                        help="Extension name of the first FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--extname2", type=str,
                        help="Extension name of the second FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--overwrite",
                        help="Overwrite output file if already exists",
                        action="store_true")
    parser.add_argument("--dtype", 
                        help="Data type of the output image (default: float32)",
                        type=str, 
                        choices=['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64',
                                 'float32', 'float64'],
                        default='float32')
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    if not args.overwrite and pathlib.Path(args.output).exists():
        print(f"Output file {args.output} already exists. Use --overwrite to overwrite it.")
        sys.exit(1)

    # compute operation
    solution = compute_operation(
        file1=args.file1, 
        file2=args.file2,
        operation=args.operation, 
        extname1=args.extname1,
        extname2=args.extname2,
        dtype=args.dtype
    )

    # save output file
    header = fits.getheader(args.file1, args.extname1)
    if header['naxis1'] != solution.shape[2] or \
       header['naxis2'] != solution.shape[1] or \
       header['naxis3'] != solution.shape[0]:
        raise ValueError("Output image dimensions do not match the input image dimensions.")
    hdu = fits.PrimaryHDU(solution.astype(args.dtype), header=header)
    add_script_info_to_fits_history(hdu.header, args)
    hdu.writeto(args.output, overwrite=args.overwrite)


if __name__ == "__main__":

    main()



#
# Copyright 2015-2022 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""
Functions to find peaks in 1D arrays and subpixel positions
"""


import numpy
from numpy import dot
from numpy.linalg import inv
from scipy.ndimage import generic_filter
from ._kernels import kernel_peak_function  # noqa


WW = dict()
WW[3] = numpy.array([[1.11022302e-16, 1.00000000e+00, 1.11022302e-16],
                     [-5.00000000e-01, 0.00000000e+00, 5.00000000e-01],
                     [5.00000000e-01, -1.00000000e+00, 5.00000000e-01]])

WW[5] = numpy.array([[-0.08571429, 0.34285714, 0.48571429, 0.34285714, -0.08571429],
                     [-0.4, -0.2, 0, 0.2, 0.4],
                     [0.57142857, -0.28571429, -0.57142857, -0.28571429, 0.57142857]])

WW[7] = numpy.array([[-9.52380952e-02, 1.42857143e-01, 2.85714286e-01, 3.33333333e-01,
                      2.85714286e-01, 1.42857143e-01, -9.52380952e-02],
                     [-3.21428571e-01, -2.14285714e-01, -1.07142857e-01, -6.79728382e-18,
                      1.07142857e-01, 2.14285714e-01, 3.21428571e-01],
                     [5.35714286e-01, 5.55111512e-17, -3.21428571e-01, -4.28571429e-01,
                      -3.21428571e-01, -2.22044605e-16, 5.35714286e-01]])

WW[9] = numpy.array([[-0.09090909, 0.06060606, 0.16883117, 0.23376623, 0.25541126,
                      0.23376623, 0.16883117, 0.06060606, -0.09090909],
                     [-0.26666667, -0.2, -0.13333333, -0.06666667, 0,
                      0.06666667, 0.13333333, 0.2, 0.26666667],
                     [0.48484848, 0.12121212, -0.13852814, -0.29437229, -0.34632035,
                      -0.29437229, -0.13852814, 0.12121212, 0.48484848]])


def _check_window_width(window_width):
    """Check `window_width` is odd and >=3"""
    if (window_width < 3) or (window_width % 2 == 0):
        raise ValueError('Window width must be an odd number and >=3')


def filter_array_margins(arr, ipeaks, window_width=5):
    _check_window_width(window_width)

    max_number = (len(arr)-1) - (window_width // 2)
    min_number = window_width // 2
    return ipeaks[(ipeaks >= min_number) & (ipeaks <= max_number)]


def find_peaks_indexes(arr, window_width=5, threshold=0.0, fpeak=0):
    """Find indexes of peaks in a 1d array.

    Note that window_width must be an odd number. The function imposes that the
    fluxes in the window_width /2 points to the left (and right) of the peak
    decrease monotonously as one moves away from the peak, except that
    it allows fpeak constant values around the peak.

    Parameters
    ----------
    arr : 1d numpy array
        Input 1D spectrum.
    window_width : int
        Width of the window where the peak must be found. This number must be
        odd.
    threshold : float
        Minimum signal in the peak (optional).
    fpeak: int
        Number of equal values around the peak

    Returns
    -------
    ipeaks : 1d numpy array (int)
        Indices of the input array arr in which the peaks have been found.


    """

    _check_window_width(window_width)

    if fpeak < 0 or fpeak + 1 >= window_width:
        raise ValueError('fpeak must be in the range 0- window_width - 2')

    kernel_peak = kernel_peak_function(threshold, fpeak)
    out = generic_filter(arr, kernel_peak, window_width, mode="reflect")
    result, =  numpy.nonzero(out)

    return filter_array_margins(arr, result, window_width)


def return_weights(window_width):
    """

    Parameters
    ----------
    window_width : int
       Odd number greater than 3, width of the window to seek for the peaks.
    Returns
    -------
    ndarray :
           Matrix needed to interpolate 'window_width' points
    """
    _check_window_width(window_width)

    try:
        return WW[window_width]
    except KeyError:
        final_ww = generate_weights(window_width)
        WW[window_width] = final_ww
        return final_ww


def generate_weights(window_width):
    """

    Parameters
    ----------
    window_width : int
                Odd number greater than 3, width of the window to seek for the peaks.

    Returns
    -------
    ndarray :
           Matrix needed to interpolate 'window_width' points
    """
    _check_window_width(window_width)

    evenly_spaced = numpy.linspace(-1, 1, window_width)
    pow_matrix = numpy.fliplr(numpy.vander(evenly_spaced, N=3))
    final_ww = dot(inv(dot(pow_matrix.T, pow_matrix)), pow_matrix.T)
    return final_ww


def refine_peaks(arr, ipeaks, window_width):
    """Refine the peak location previously found by find_peaks_indexes

    Parameters
    ----------
    arr : 1d numpy array, float
        Input 1D spectrum.
    ipeaks : 1d numpy array (int)
        Indices of the input array arr in which the peaks were initially found.
    window_width : int
        Width of the window where the peak must be found.

    Returns
    -------
     xc, yc: tuple
        X-coordinates in which the refined peaks have been found,
        interpolated Y-coordinates

    """
    _check_window_width(window_width)

    step = window_width // 2

    ipeaks = filter_array_margins(arr, ipeaks, window_width)

    winoff = numpy.arange(-step, step+1, dtype='int')
    peakwin = ipeaks[:, numpy.newaxis] + winoff
    ycols = arr[peakwin]

    ww = return_weights(window_width)

    coff2 = numpy.dot(ww, ycols.T)

    uc = -0.5 * coff2[1] / coff2[2]
    yc = coff2[0] + uc * (coff2[1] + coff2[2] * uc)
    xc = ipeaks + 0.5 * (window_width-1) * uc

    return xc, yc



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compute array sizes in a human-readable format"""

from astropy.units import Unit


def bytes_to(size_in_bytes, to, bsize=1024):
    if to != 'k':
        to = to.upper()
    s = {'k': 1, 'M': 2, 'G': 3, 'T': 4, 'P': 5, 'E': 6}
    if to in s:
        return size_in_bytes / (bsize ** s[to])
    else:
        raise ValueError(f'Unexpected {to=}. Invalid size unit.')


def array_size_32bits(array):
    """Return 32bits-array size in a human-readable format"""
    size_in_bytes = array.__sizeof__()
    if array.dtype == 'float64':
        size_in_bytes /= 2
    elif array.dtype == 'float32':
        pass
    else:
        raise ValueError(f'Invalid {array.dtype=}')
    for s in 'EPTGMk':
        size = bytes_to(size_in_bytes, to=s)
        if size > 1:
            return size * Unit(f'{s}byte')
    raise ValueError(f'Size: {size_in_bytes} bytes is too large!')


def array_size_8bits(array):
    """Return 8bits-array size in a human-readable format"""
    size_in_bytes = array.__sizeof__()
    if array.dtype == 'float64':
        size_in_bytes /= 8
    elif array.dtype == 'float32':
        size_in_bytes /= 4
    elif array.dtype == 'uint16':
        size_in_bytes /= 2
    elif array.dtype == 'uint8':
        pass
    else:
        raise ValueError(f'Invalid {array.dtype=}')
    for s in 'EPTGMk':
        size = bytes_to(size_in_bytes, to=s)
        if size > 1:
            return size * Unit(f'{s}byte')
    raise ValueError(f'Size: {size_in_bytes} bytes is too large!')



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Plot histogram using the matplotlib step function."""
import numpy as np


def plot_hist_step(ax, bins, h, label=None,
                   linestyle='-',
                   color='C0', alpha=1.0,
                   fill_color=None, fill_alpha=0.4,
                   left_border=True, right_border=True):
    """Plot a histogram using the step function.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot on.
    bins : array_like
        The bin edges for the histogram.
    h : array_like
        The histogram values.
    label : str, optional
        Label for the histogram. Default is None.
    linestyle : str, optional
        Line style for the step line. Default is '-'.
    color : str, optional
        Color of the step line. Default is 'C0'.
    alpha : float, optional
        Transparency of the step line. Default is 1.0.
    fill_color : str, optional
        Color to fill the area under the step line. Default is None (no fill).
    fill_alpha : float, optional
        Transparency of the filled area. Default is 0.4.
    left_border : bool, optional
        If True, draw a vertical line at the left edge of the first bin.
        Default is True.
    right_border : bool, optional
        If True, draw a vertical line at the right edge of the last bin.
        Default is True.
    """
    xdum = (bins[:-1] + bins[1:]) / 2
    ax.step(xdum, h, where='mid', color=color, linestyle=linestyle, label=label)
    if left_border:
        ax.plot([bins[0], bins[0], xdum[0]], [0, h[0], h[0]], alpha=alpha, color=f'{color}', linestyle=linestyle)
    if right_border:
        ax.plot([xdum[-1], bins[-1], bins[-1]], [h[-1], h[-1], 0], alpha=alpha, color=f'{color}', linestyle=linestyle)
    if fill_color is not None:
        xfill = xdum
        hfill = h
        if left_border:
            xfill = np.concatenate((np.array([bins[0]]), xdum))
            hfill = np.concatenate((np.array([h[0]]), h))
        if right_border:
            xfill = np.concatenate((xdum, np.array([bins[-1]])))
            hfill = np.concatenate((h, np.array([h[-1]])))
        ax.fill_between(xfill, hfill, step='mid', alpha=fill_alpha, color=f'{fill_color}')



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np


def find_pix_borders(sp, sought_value):
    """Find useful region of a given spectrum

    Detemine the useful region of a given spectrum by skipping
    the initial (final) pixels with values equal to 'sought_value'.

    Parameters
    ----------
    sp : 1D numpy array
        Input spectrum.
    sought_value : int, float, bool
        Pixel value that indicate missing data in the spectrum.

    Returns
    -------
    jmin, jmax : tuple (integers)
        Valid spectrum region (in array coordinates, from 0 to
        NAXIS1 - 1). If the values of all the pixels in the spectrum
        are equal to 'sought_value', the returned values are jmin=-1
        and jmax=naxis1.

    """

    if sp.ndim != 1:
        raise ValueError('Unexpected number of dimensions:', sp.ndim)
    naxis1 = len(sp)

    jborder_min = -1
    jborder_max = naxis1

    # only spectra with values different from 'sought_value'
    if not np.all(sp == sought_value):
        # left border
        while True:
            jborder_min += 1
            if sp[jborder_min] != sought_value:
                break
        # right border
        while True:
            jborder_max -= 1
            if sp[jborder_max] != sought_value:
                break

    return jborder_min, jborder_max


def fix_pix_borders(image2d, nreplace, sought_value, replacement_value):
    """Replace a few pixels at the borders of each spectrum.

    Set to 'replacement_value' 'nreplace' pixels at the beginning (at
    the end) of each spectrum just after (before) the spectrum value
    changes from (to) 'sought_value', as seen from the image borders.

    Parameters
    ----------
    image2d : numpy array
        Initial 2D image.
    nreplace : int
        Number of pixels to be replaced in each border.
    sought_value : int, float, bool
        Pixel value that indicates missing data in the spectrum.
    replacement_value : int, float, bool
        Pixel value to be employed in the 'nreplace' pixels.

    Returns
    -------
    image2d : numpy array
        Final 2D image.

    """

    # input image size
    naxis2, naxis1 = image2d.shape

    for i in range(naxis2):
        # only spectra with values different from 'sought_value'
        jborder_min, jborder_max = find_pix_borders(
            image2d[i, :],
            sought_value=sought_value
        )

        # left border
        if jborder_min != -1:
            j1 = jborder_min
            j2 = min(j1 + nreplace, naxis1)
            image2d[i, j1:j2] = replacement_value
        # right border
        if jborder_max != naxis1:
            j2 = jborder_max + 1
            j1 = max(j2 - nreplace, 0)
            image2d[i, j1:j2] = replacement_value

    return image2d


def define_mask_borders(image2d, sought_value, nadditional=0):
    """Generate mask avoiding undesired values at the borders.

    Set to True image borders with values equal to 'sought_value'

    Parameters
    ----------
    image2d : numpy array
        Initial 2D image.
    sought_value : int, float, bool
        Pixel value that indicates missing data in the spectrum.
    nadditional : int
        Number of additional pixels to be masked at each border.

    Returns
    -------
    mask2d : numpy array
        2D mask.
    borders : list of tuples
        List of tuples (jmin, jmax) with the border limits (in array
        coordinates) found by find_pix_borders.

    """

    # input image size
    naxis2, naxis1 = image2d.shape

    # initialize mask
    mask2d = np.zeros((naxis2, naxis1), dtype=bool)

    # initialize list to store borders
    borders = []

    for i in range(naxis2):
        # only spectra with values different from 'sought_value'
        jborder_min, jborder_max = find_pix_borders(
            image2d[i, :],
            sought_value=sought_value
        )
        borders.append((jborder_min, jborder_max))
        if (jborder_min, jborder_max) != (-1, naxis1):
            if jborder_min != -1:
                j1 = 0
                j2 = jborder_min + nadditional + 1
                mask2d[i, j1:j2] = True
            if jborder_max != naxis1:
                j1 = jborder_max - nadditional
                j2 = naxis1
                mask2d[i, j1:j2] = True

    return mask2d, borders



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Numina data processing system."""


import logging

from ._version import version

__version__ = version

# Top level NullHandler
logging.getLogger("numina").addHandler(logging.NullHandler())



#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from datetime import datetime
import itertools
import json
import pathlib
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable

import attrs
from dateutil.parser import isoparse
import importlib_resources
import numina.util.objimport

from .elements import SetupBlock, PropertiesBlock, ElementEnum
from .generic import CG
from .configorigin import ElementOrigin
from .property import (
    PropertyEntry,
    PropertyProxy,
    PropertyModOneOf,
    PropertyModLimits,
    PropertyBase,
)
from ..keydef import KeyDefinition


@attrs.define
class ComponentCollection:
    dirname = attrs.field()
    paths = attrs.field()


def load_resources_dir(dirname):
    raise NotImplementedError
    # return ComponentCollection(dirname, valid_paths)


def load_comp_store(comp_collection: ComponentCollection) -> dict:
    """

    Parameters
    ----------
    comp_collection : ComponentCollection

    Returns
    -------
    dict
    """
    comp_store = {}
    for entry in comp_collection.dirname.iterfiles():
        if entry.name in comp_collection.paths:
            with open(entry) as fd:
                cont = json.load(fd)
                cont["origin"] = ElementOrigin.from_dict(cont)
                comp_store[entry.name] = cont
    return comp_store


def load_panoply_store(sys_drps=None, defpath=None) -> dict:
    if defpath is None:
        file_paths = []
    else:
        file_paths = [defpath]

    pkg_paths = []
    # DRP paths
    if sys_drps is not None:
        for drp in sys_drps.drps.values():
            pkg_paths.append(drp.profiles)

    return load_paths_store(pkg_paths, file_paths)


def load_paths_store(pkg_paths=None, file_paths=None) -> dict:
    """

    Parameters
    ----------
    pkg_paths
    file_paths

    Returns
    -------
    dict
    """

    comp_store = {}
    # Prepare file paths
    if file_paths is None:
        file_paths = []
    if pkg_paths is None:
        pkg_paths = []

    paths1 = [pathlib.Path(fpath) for fpath in file_paths]
    paths2 = [importlib_resources.files(ppath) for ppath in pkg_paths]

    for path in itertools.chain(paths1, paths2):
        for obj in path.iterdir():
            if obj.suffix == ".json":
                with open(obj) as fd:
                    cont = json.load(fd)
                    cont["origin"] = ElementOrigin.from_dict(cont)
                    comp_store[obj.name] = cont

    return comp_store


def find_instrument(
    comp_store, keyval: str, date: str | datetime, by_key="name"
) -> dict[str, Any]:
    """
    Find instrument in the component collection

    Raises
    ------
    ValueError
        If there is no instrument
    """
    return find_element(
        comp_store, ElementEnum.ELEM_INSTRUMENT, keyval, date, by_key=by_key
    )


def find_element(
    comp_store, etype: ElementEnum, keyval: str, date: str | datetime, by_key="name"
) -> dict[str, Any]:
    """
    Find component of the given type in the component collection

    Raises
    ------
    ValueError
        If there is no component
    """
    if isinstance(date, str):
        datet = isoparse(date)
    else:
        datet = date

    element_name = ElementEnum.to_str(etype)

    for key, val in comp_store.items():
        if (keyval == val[by_key]) and (val["type"] == element_name):
            if val["origin"].is_valid_date(datet):
                return val
            else:
                # print('date not valid', datet, val['origin'].date_start, val['origin'].date_end)
                pass
    else:
        raise ValueError(f"Not found {element_name} {by_key}={keyval} for date={date}")


def assembly_instrument(
    comp_store, keyval: str, date: str | datetime, by_key: str = "name"
) -> CG:
    """
    Assembly an instrument configuration object.

    Create an instrument object from a store of configurations using
    either the UUID of the configuration or the date of the configuration

    Parameters
    ----------
    comp_store : dict
    keyval : str
    date : str or datetime
    by_key : str, optional

    Returns
    -------
    InstrumentGeneric
        an instrument configuration
    """
    return assembly_element(
        comp_store, ElementEnum.ELEM_INSTRUMENT, keyval, date, by_key=by_key
    )


def assembly_component(
    comp_store, keyval: str, date: str | datetime, dest=None, by_key="name"
) -> CG:
    """
    Assembly an instrument configuration object.

    Create an instrument object from a store of configurations using
    either the UUID of the configuration or the date of the configuration

    Parameters
    ----------
    comp_store : dict
    keyval : str
    date : str or datetime
    dest: str, optional
    by_key : str, optional

    Returns
    -------
    ComponentGeneric
        an instrument configuration
    """
    return assembly_element(
        comp_store, ElementEnum.ELEM_COMPONENT, keyval, date, dest=dest, by_key=by_key
    )


def assembly_property(
    comp_store, keyval: str, date: str | datetime, dest=None, by_key="name"
) -> PropertiesBlock:
    return assembly_element(
        comp_store, ElementEnum.ELEM_PROPERTIES, keyval, date, dest=dest, by_key=by_key
    )


def assembly_setup(
    comp_store, keyval: str, date: str | datetime, dest: str = None, by_key="name"
) -> SetupBlock:
    return assembly_element(
        comp_store, ElementEnum.ELEM_SETUP, keyval, date, dest=dest, by_key=by_key
    )


def assembly_element(
    comp_store,
    etype: ElementEnum,
    tmpl: str,
    date: str | datetime,
    dest: str | None = None,
    by_key="name",
):
    """
    Assembly a given element in the object.
    Parameters
    ----------
    comp_store : dict
    etype: ElementEnum
    tmpl
    date : str or datetime
    dest
    by_key : str, optional

    Returns
    -------
    ComponentGeneric or SetupGeneric or PropertiesGeneric

    Raises
    ------
    ValueError is there is a problem during element construction
    """

    somed = find_element(comp_store, etype, tmpl, date, by_key=by_key)

    if dest is None:
        dest_name = somed["name"]
    else:
        dest_name = dest
    # print('dest', dest_name, somed['name'], somed['uuid'])
    if "class" in somed:
        class_name = somed["class"]
        clss = numina.util.objimport.import_object(class_name)
    else:
        class_name = somed["name"]
        clss = ElementEnum.to_class(etype)

    setup_block = somed.get("setup", [])
    setup_obj = process_setup(comp_store, somed["name"], setup_block, date)

    prop_block = somed.get("properties", [])
    prop_objects = process_properties(comp_store, prop_block, date)
    # print('prop', prop_objects)

    iml = clss.from_component(
        class_name,
        dest_name,
        origin=somed["origin"],
        properties=prop_objects,
        setup=setup_obj,
    )
    # if len(setup_obj.values) > 0:
    #    print(setup_block, Klss)
    comp_block = somed.get("components", [])
    comp_objects = process_components(comp_store, comp_block, date)

    # print('comp', comp_objects)
    for comp in comp_objects:
        comp.set_parent(iml)

    return iml


def process_setup(
    comp_store,
    setup_id: str,
    setup_block: "Iterable[dict[str, Any]]",
    date: str | datetime,
) -> SetupBlock:
    setup_objects = {}
    for entry in setup_block:
        match entry:
            case {"values": values}:
                cid = entry.get("id", "values")
                setup_objects[cid] = values
            case {"name": name, "id": cid}:
                res = assembly_setup(comp_store, name, date, dest=cid, by_key="name")
                setup_objects[res.name] = res
            case {"uuid": name, "id": cid}:
                res = assembly_setup(comp_store, name, date, dest=cid, by_key="uuid")
                setup_objects[res.name] = res

            case _:
                raise ValueError(f"error in setup: {entry}")

    setup_obj = SetupBlock(setup_id)
    for key, val in setup_objects.items():
        if isinstance(val, SetupBlock):
            setup_obj.values.update(val.values)
        else:
            setup_obj.values[key] = val

    return setup_obj


def process_properties(
    comp_store, prop_block, date: str | datetime
) -> dict[str, PropertyBase]:
    prop_objects: dict[str, PropertyBase] = {}
    for entry in prop_block:
        key = entry["id"]
        match entry:
            case {"values": values, "depends": depends}:
                confe = PropertyEntry(values, depends)
            case {"one_of": one_of, **rest}:
                if len(one_of) < 1:
                    raise ValueError(f'"one_of must have length >= 1: {entry}')
                default = rest.get("default", one_of[0])
                # FIXME: default is stored twice
                mapping = process_key_def(rest, default=default)
                confe = PropertyModOneOf(one_of, default=default, mapping=mapping)
            case {"limits": limits, **rest}:
                default = rest.get("default", 0)
                # FIXME: default is stored twice
                mapping = process_key_def(rest, default=default)
                confe = PropertyModLimits(limits, default=default, mapping=mapping)
            case {"references": ref_name}:
                confe = PropertyProxy(ref_name)
            case {"name": e_name}:
                res = assembly_property(comp_store, e_name, date)
                confe = res.properties[key]
            case {"uuid": e_uuid}:
                res = assembly_property(comp_store, e_uuid, date, by_key="uuid")
                confe = res.properties[key]
            case _:
                raise ValueError(f"error in properties: {entry}")

        prop_objects[key] = confe
    return prop_objects


def process_components(comp_store, comp_block, date: str | datetime) -> list[CG]:
    # Add components when device is initialised
    # print('load components')
    comp_objects = []
    for entry in comp_block:
        # Use 'id' as name, if present
        cid = entry.get("id")
        if "name" in entry:
            key = "name"
        elif "uuid" in entry:
            key = "uuid"
        else:
            raise ValueError(f"error in components: {entry}")

        res = assembly_component(comp_store, entry[key], date, dest=cid, by_key=key)
        comp_objects.append(res)
    return comp_objects


def process_key_def(rest: dict[str, Any], default) -> KeyDefinition | None:
    key = rest.get("key")
    ext = rest.get("ext")
    if key is not None:
        return KeyDefinition(key, ext=ext, default=default)
    return None



#
# Copyright 2011-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DRP system wide initialization"""


from .drpsystem import DrpSystem

_system_drps = None


def get_system_drps():
    """Load all compatible DRPs in the system"""
    global _system_drps
    if _system_drps is None:
        _system_drps = DrpSystem()
        _system_drps.load()

    return _system_drps



#
# Copyright 2016-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""A representation of a hardware device"""

import inspect
import typing

from .device import DeviceBase


class HWDevice(DeviceBase):

    def init_config_info(self) -> dict[str, typing.Any]:
        return dict(name=self.name)

    def get_property_names(self) -> list[str]:
        result = []
        for key in dir(self.__class__):
            if key.startswith("_"):
                continue
            attr = getattr(self.__class__, key, None)
            if inspect.isdatadescriptor(attr):
                result.append(key)
        return result



"""Combine several FITS images listed in a TXT file"""

import argparse
import datetime
import sys

from astropy.io import fits
import numpy as np

from numina.array.display.fileinfo import list_fileinfo_from_txt
from numina.processing.combine import combine
from .subsets_of_fileinfo_from_txt import subsets_of_fileinfo_from_txt


def compute_abba_result(list_of_fileinfo, outfile, extnum=0, noheader=False,
                        save_partial=False, debugplot=0):
    """Compute A-B-(B-A).

    Note that 'list_of_fileinfo' must contain 4 file names.

    Parameters
    ----------
    list_of_fileinfo : list of strings
        List of the for files corresponding to the ABBA observation.
    outfile : string
        Base name for the output FITS file name.
    extnum : int
        Extension number to be read (default value 0 = primary extension).
    noheader : bool
        If True, the header of the first image in the ABBA sequence is
        not copied in the resulting image(s).
    save_partial : bool
        If True, the partial (A-B) and -(B-A) images are also saved.
    debugplot : integer or None
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    """

    # check number of images
    if len(list_of_fileinfo) != 4:
        raise ValueError("Unexpected number of ABBA files: " +
                         str(len(list_of_fileinfo)))

    # avoid PyCharm warnings
    # (local variable might be referenced before assignment)
    naxis1 = 0
    naxis2 = 0
    image_header = None

    # check image dimensions
    for i in range(4):
        hdulist = fits.open(list_of_fileinfo[i].filename)
        if i == 0:
            image_header = hdulist[extnum].header
            naxis1 = image_header['naxis1']
            naxis2 = image_header['naxis2']
            hdulist.close()
        else:
            image_header_ = hdulist[extnum].header
            naxis1_ = image_header_['naxis1']
            naxis2_ = image_header_['naxis2']
            hdulist.close()
            if naxis1 != naxis1_ or naxis2 != naxis2_:
                print('>>> naxis1, naxis2..:', naxis1, naxis2)
                print('>>> naxis1_, naxis2_:', naxis1_, naxis2_)
                raise ValueError("Image dimensions do not agree!")

    # initialize outpuf array(s)
    if save_partial:
        result_ini = np.zeros((naxis2, naxis1), dtype=float)
        result_end = np.zeros((naxis2, naxis1), dtype=float)
    else:
        result_ini = None
        result_end = None
    result = np.zeros((naxis2, naxis1), dtype=float)

    # read the four images and compute arithmetic combination
    for i in range(4):
        if abs(debugplot) >= 10:
            print('Reading ' + list_of_fileinfo[i].filename + '...')
        hdulist = fits.open(list_of_fileinfo[i].filename)
        image2d = hdulist[extnum].data.astype(float)
        hdulist.close()
        if i == 0:
            if save_partial:
                result_ini += image2d
            result += image2d
        elif i == 1:
            if save_partial:
                result_ini -= image2d
            result -= image2d
        elif i == 2:
            if save_partial:
                result_end -= image2d
            result -= image2d
        elif i == 3:
            if save_partial:
                result_end += image2d
            result += image2d
        else:
            raise ValueError('Unexpected image number: ' + str(i))

    # save results
    if noheader:
        image_header = None
    if save_partial:
        if abs(debugplot) >= 10:
            print("==> Generating output file: " + outfile + "_sub1.fits...")
        hdu = fits.PrimaryHDU(result_ini.astype(float), image_header)
        hdu.writeto(outfile + '_sub1.fits', overwrite=True)
        if abs(debugplot) >= 10:
            print("==> Generating output file: " + outfile + "_sub2.fits...")
        hdu = fits.PrimaryHDU(result_end.astype(float), image_header)
        hdu.writeto(outfile + '_sub2.fits', overwrite=True)
    if abs(debugplot) >= 10:
        print("==> Generating output file: " + outfile + ".fits...")
    hdu = fits.PrimaryHDU(result.astype(float), image_header)
    hdu.writeto(outfile + '.fits', overwrite=True)


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()
    parser.add_argument("input_list",
                        help="TXT file with list of images to be coadded")
    # if output_fits_filename == "@" the script expects a line in
    # the file input_list of the form:
    # @ <number> <label>
    # where <number> is the number of images to be coadded and
    # <label> is the name of the output file
    parser.add_argument('output_fits_filename',
                        help='filename of output FITS image, or @ symbol')
    parser.add_argument('--method',
                        help='Combination method: sum (default), ' +
                             'mean, median, sigmaclip, abba, abba_partial',
                        default='sum',
                        type=str,
                        choices=['sum', 'mean', 'median', 'sigmaclip',
                                 'abba', 'abba_partial'])
    parser.add_argument('--method_kwargs',
                        help='Arguments for method sigmaclip; must be a'
                             'Python dictionary between double quotes, e.g.: '
                             '"{' + "'high': 2.5, 'low': 2.5}" + '")',
                        type=str)
    parser.add_argument('--extnum',
                        help='Extension number in input files (note that ' +
                             'first extension is 1 = default value)',
                        default=1, type=int)
    parser.add_argument('--noheader',
                        help='Do not include header of first image in ' +
                             'outpuf file(s)',
                        action='store_true')
    parser.add_argument('--noclobber',
                        help='Avoid overwriting existing file',
                        action='store_true')
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                             " (default=12)",
                        default=12, type=int,
                        choices=[0, 1, 2, 10, 11, 12, 21, 22])
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # first extension is number 1 for the user
    extnum = args.extnum - 1

    if args.output_fits_filename == "@":
        dict_of_fileinfo = subsets_of_fileinfo_from_txt(args.input_list)
    else:
        list_of_fileinfo = list_fileinfo_from_txt(args.input_list)
        dict_of_fileinfo = {}
        dict_of_fileinfo[0] = {
            'label': args.output_fits_filename,
            'list_of_fileinfo': list_of_fileinfo
        }

    number_of_output_files = len(dict_of_fileinfo)

    # main loop in number of output files
    for iout in range(number_of_output_files):
        # list of files to be combined
        tmpdict = dict_of_fileinfo[iout]
        output_fits_filename = tmpdict['label']
        list_of_files = [tmp.filename for tmp in tmpdict['list_of_fileinfo']]

        # number of files to be combined
        number_of_files = len(list_of_files)

        # declare auxiliary arrays to store image basic parameters
        naxis1 = np.zeros(number_of_files, dtype=int)
        naxis2 = np.zeros(number_of_files, dtype=int)

        # read basic parameters for all the images
        for i in range(number_of_files):
            infile = list_of_files[i]
            hdulist = fits.open(infile)
            image_header = hdulist[extnum].header
            hdulist.close()
            naxis1[i] = image_header['naxis1']
            naxis2[i] = image_header['naxis2']

        # check that NAXIS1 is the same for all the images
        naxis1_comp = np.repeat(naxis1[0], number_of_files)
        if not np.allclose(naxis1, naxis1_comp):
            raise ValueError("NAXIS1 values are different")

        # check that NAXIS2 is the same for all the images
        naxis2_comp = np.repeat(naxis2[0], number_of_files)
        if not np.allclose(naxis2, naxis2_comp):
            raise ValueError("NAXIS2 values are different")

        image_header_first_frame = None  # avoid PyCharm warning
        image2d = None                   # avoid PyCharm warning

        if args.method in ['sum', 'mean']:
            image2d = np.zeros((naxis2[0], naxis1[0]), dtype=float)
            # add all the individual images
            for i in range(number_of_files):
                infile = list_of_files[i]
                print("<--" + infile + " (image " + str(i + 1) + " of " +
                      str(number_of_files) + ')')
                hdulist = fits.open(infile)
                data = hdulist[extnum].data
                if i == 0:
                    image_header_first_frame = hdulist[extnum].header
                hdulist.close()
                image2d += data
            # average result when required
            if args.method == 'mean':
                image2d /= number_of_files
        elif args.method == 'sigmaclip':
            list_data = []
            for i in range(number_of_files):
                infile = list_of_files[i]
                print("<--" + infile + " (image " + str(i + 1) + " of " +
                      str(number_of_files) + ')')
                hdulist = fits.open(infile)
                data = hdulist[extnum].data
                if i == 0:
                    image_header_first_frame = hdulist[extnum].header
                hdulist.close()
                list_data.append(data)
            if args.method_kwargs is None:
                image2d = combine.sigmaclip(list_data)[0]
            else:
                method_kwargs = eval(args.method_kwargs)
                image2d = combine.sigmaclip(list_data, **method_kwargs)[0]
        elif args.method == 'median':
            # declare temporary cube to store all the images
            image3d = np.zeros((number_of_files, naxis2[0], naxis1[0]),
                               dtype=float)
            # read all the individual images
            for i in range(number_of_files):
                infile = list_of_files[i]
                print("<--" + infile + " (image " + str(i + 1) + " of " +
                      str(number_of_files) + ')')
                hdulist = fits.open(infile)
                data = hdulist[extnum].data
                if i == 0:
                    image_header_first_frame = hdulist[extnum].header
                hdulist.close()
                image3d[i, :, :] += data
            # compute median
            image2d = np.median(image3d, axis=0)
        elif args.method in ['abba', 'abba_partial']:
            save_partial = False
            if args.method == 'abba_partial':
                save_partial = True
            compute_abba_result(tmpdict['list_of_fileinfo'],
                                output_fits_filename,
                                extnum=extnum,
                                noheader=args.noheader,
                                save_partial=save_partial,
                                debugplot=args.debugplot)
        else:
            raise ValueError('Unexpected combination method!')

        # save results (except for ABBA combinations, which results
        # have already been saved)
        if args.method in ['sum', 'mean', 'median', 'sigmaclip']:
            if args.noheader:
                image_header = fits.Header()
            else:
                image_header = image_header_first_frame
            if abs(args.debugplot) >= 10:
                print("==> Generating output file: " + output_fits_filename +
                      "...")
            image_header.add_history("---")
            image_header.add_history("Image generated using:")
            image_header.add_history(" ".join(sys.argv))
            image_header.add_history("---")
            t_str = datetime.datetime.now(datetime.timezone.utc).isoformat()
            image_header.add_history(f'Combination time: {t_str}')
            image_header.add_history(f"Contents of {args.input_list} file:")
            for i in range(number_of_files):
                image_header.add_history(list_of_files[i])
            image_header.add_history("---")
            hdu = fits.PrimaryHDU(image2d.astype(float), image_header)
            hdu.writeto(output_fits_filename, overwrite=(not args.noclobber))


if __name__ == "__main__":

    main()



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import unittest
import numpy

import numina.array.combine as c


class QuantileClipTestCase(unittest.TestCase):
    """Test case for the quantileclip rejection method."""

    def setUp(self):
        self.nimages = 10
        self.data = [numpy.ones((2, 2))] * self.nimages

    def testBasic0(self):
        """Test ValueError is raised if fraction of points rejected is < 0.0"""
        self.assertRaises(ValueError, c.quantileclip, self.data, fclip=-0.01)

    def testBasic1(self):
        """Test ValueError is raised if fraction of points rejected is > 0.4"""
        self.assertRaises(ValueError, c.quantileclip, self.data, fclip=0.41)

    def testBasic2(self):
        """Test integer rejections"""
        r = c.quantileclip(self.data, fclip=0.0)
        for v in r[0].flat:
            self.assertEqual(v, 1)
        for v in r[1].flat:
            self.assertEqual(v, 0)
        for v in r[2].flat:
            self.assertEqual(v, 10)

        r = c.quantileclip(self.data, fclip=0.1)
        for v in r[0].flat:
            self.assertEqual(v, 1)
        for v in r[1].flat:
            self.assertEqual(v, 0)
        for v in r[2].flat:
            self.assertEqual(v, 8)

        r = c.quantileclip(self.data, fclip=0.2)
        for v in r[0].flat:
            self.assertEqual(v, 1)
        for v in r[1].flat:
            self.assertEqual(v, 0)
        for v in r[2].flat:
            self.assertEqual(v, 6)

    def testBasic3(self):
        """Test simple fractional rejections"""
        r = c.quantileclip(self.data, fclip=0.23)
        for v in r[0].flat:
            self.assertEqual(v, 1)
        for v in r[1].flat:
            self.assertEqual(v, 0)
        for v in r[2].flat:
            self.assertAlmostEqual(v, 5.0)

    def testBasic4(self):
        """Test complex fractional rejections"""
        data = [numpy.array([i]) for i in range(10)]

        r = c.quantileclip(data, fclip=0.0)
        for v in r[0].flat:
            self.assertAlmostEqual(v, 4.5)
        # for v in r[1].flat:
        #    self.assertAlmostEqual(v, 0.9166666666)
        for v in r[2].flat:
            self.assertAlmostEqual(v, 10)

        r = c.quantileclip(data, fclip=0.1)
        for v in r[0].flat:
            self.assertAlmostEqual(v, 4.5)
        # for v in r[1].flat:
        #    self.assertAlmostEqual(v, 0.75)
        for v in r[2].flat:
            self.assertAlmostEqual(v, 8)

        r = c.quantileclip(data, fclip=0.2)
        for v in r[0].flat:
            self.assertAlmostEqual(v, 4.5)
        # for v in r[1].flat:
        #    self.assertAlmostEqual(v, 0.58333333333333337)
        for v in r[2].flat:
            self.assertAlmostEqual(v, 6)

        r = c.quantileclip(data, fclip=0.09)
        for v in r[0].flat:
            self.assertAlmostEqual(v, 4.5)
        # for v in r[1].flat:
        #    self.assertAlmostEqual(v, 0.76666666666666672)
        for v in r[2].flat:
            self.assertAlmostEqual(v, 9.0)

    def testResults5(self):
        """Test deviant points are ignored"""
        data = [numpy.array([1.0]) for _ in range(22)]
        data[0][0] = 89
        data[12][0] = -89

        r = c.quantileclip(data, fclip=0.15)
        for v in r[0].flat:
            self.assertAlmostEqual(v, 1.0)
        for v in r[1].flat:
            self.assertAlmostEqual(v, 0.0)
        for v in r[2].flat:
            self.assertAlmostEqual(v, 15.0)
